#include "libxml/globals.h"
#include "libxml/valid.h"
#include "libxml/tree.h"

#include <vector>
#include <string>
#include <map>
#include <sstream>
#include <fstream>
#include <iostream>

#include <globus_common.h>
#include <globus_rsl.h>

#include "ngui.h"
#include "ngdata.h"
#include "ngexec.h"

#define SI2000PERMHZ 28.5

using namespace std;

struct executorData {
  string jobDefID;
  string jobName;
  string jobExeID;
};

struct jobExeData {
  string jobDefID;
  string jobName;
  string jobExeID;
  string infoExecutor;
  string action;
};

struct cpuConsumption {
  string count;
  string unit;
};

struct diskConsumption {
  string count;
  string unit;
};

struct minimumRAM {
  string count;
  string unit;
};

struct fileInfo {
  string stream;
  string LFN;
  string type;
  string logCol;
  vector <string> turlList;
  vector <string> SEList;
};

struct actualPar {
  string name;
  int position;
  string type;
  string value;
};

struct jobDef {
  string jobDefID;
  string transImpl;
  string transUses;
  string transVersion;
  string jobName;
  cpuConsumption CPU;
  diskConsumption disk;
  minimumRAM ram;
  string ipconnectivity;
  vector <fileInfo> jobLogs;
  vector <fileInfo> jobInputs;
  vector <fileInfo> jobOutputs;
  vector <actualPar> jobPars;
};

static void GetLine( string& s, string& line ) {
  int i = s.find('\n');
  if( i == string::npos ) return;

  line = s.substr( 0, i );
  s = s.substr( i+1, s.size()-1 );
}


static string GetEnv (const string & env) {

  const char * temp = getenv (env.c_str());
  return temp ? temp : "";
}


static int Minutes (const string & period) {

  int minutes = 0;

  int pos = string::npos;
  int len;

  for (int i = 0; i != period.length(); i++) {
    if (isdigit (period[i])) {
      if (pos == string::npos) {pos = i; len = 0;}
      len ++;
    }
    else if (pos != string::npos) {
      switch (period[i]) {
      case 'w':
      case 'W':
        minutes += atoi (period.substr (pos, len).c_str()) * 60 * 24 * 7;
        pos = string::npos;
        break;
      case 'd':
      case 'D':
        minutes += atoi (period.substr (pos, len).c_str()) * 60 * 24;
        pos = string::npos;
        break;
      case 'h':
      case 'H':
        minutes += atoi (period.substr (pos, len).c_str()) * 60;
        pos = string::npos;
        break;
      case 'm':
      case 'M':
        minutes += atoi (period.substr (pos, len).c_str());
        pos = string::npos;
        break;
      case ' ':
        break;
      default:
        return -1;
        break;
      }
    }
  }
  if (pos != string::npos)
    minutes += atoi (period.substr (pos, len).c_str());
  return minutes;
}


int FindRelation (const string & attr, globus_rsl_t ** relation,
		  globus_rsl_t * axrsl) {

  if (globus_rsl_is_boolean (axrsl)) {
    globus_list_t * rlist = globus_rsl_boolean_get_operand_list (axrsl);
    while (!globus_list_empty (rlist)) {
      if (FindRelation (attr, relation,
			(globus_rsl_t *) globus_list_first (rlist)))
	return 1;
      rlist = globus_list_rest (rlist);
    }
    return 0;
  }
  else if (globus_rsl_is_relation (axrsl)) {
    if (globus_rsl_is_relation_attribute_equal (axrsl,
						(char *) attr.c_str())) {
      if (*relation) {
	cerr << "Error: Multiple definitions of XRSL attribute \"" << attr
	     << "\"" << endl;
	return 1;
      }
      *relation = axrsl;
    }
    return 0;
  }
  else {
    cerr << "Error: Unexpected XRSL token" << endl;
    return 1;
  }
}


char * numJobsWanted (char * xml) {

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

  xmlNodePtr xmlnode1 = xmlroot->xmlChildrenNode;

  cpuConsumption cpu;
  diskConsumption disk;
  minimumRAM ram;
  string transUses;
  string ipConnectivity;

  while (xmlnode1) {
    if (strcmp ((const char *) xmlnode1->name, "numJobsWanted") == 0) {
      xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;
      while (xmlnode2) {
	if (strcmp ((const char *) xmlnode2->name, "minimumResources") == 0) {
	  xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	  while (xmlnode3) {
	    if (strcmp ((const char *) xmlnode3->name, "transUses") == 0) {
	      if (xmlnode3->xmlChildrenNode)
		transUses = (const char *) xmlnode3->xmlChildrenNode->content;
	    }
	    if (strcmp ((const char *) xmlnode3->name, "cpuConsumption") == 0) {
	      xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
	      while (xmlnode4) {
		if (strcmp ((const char *) xmlnode4->name, "count") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    cpu.count = (const char *) xmlnode4->xmlChildrenNode->content;
		if (strcmp ((const char *) xmlnode4->name, "unit") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    cpu.unit = (const char *) xmlnode4->xmlChildrenNode->content;
		xmlnode4 = xmlnode4->next;
	      }
	    }
	    if (strcmp ((const char *) xmlnode3->name, "diskConsumption") == 0) {
	      xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
	      while (xmlnode4) {
		if (strcmp ((const char *) xmlnode4->name, "count") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    disk.count = (const char *) xmlnode4->xmlChildrenNode->content;
		if (strcmp ((const char *) xmlnode4->name, "unit") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    disk.unit = (const char *) xmlnode4->xmlChildrenNode->content;
		xmlnode4 = xmlnode4->next;
	      }
	    }
	    if (strcmp ((const char *) xmlnode3->name, "ipConnectivity") == 0) {
	      if (xmlnode3->xmlChildrenNode)
		ipConnectivity = (const char *) xmlnode3->xmlChildrenNode->content;
	    }
	    if (strcmp ((const char *) xmlnode3->name, "mimimumRAM") == 0) {
	      xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
	      while (xmlnode4) {
		if (strcmp ((const char *) xmlnode4->name, "count") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    ram.count = (const char *) xmlnode4->xmlChildrenNode->content;
		if (strcmp ((const char *) xmlnode4->name, "unit") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    ram.unit = (const char *) xmlnode4->xmlChildrenNode->content;
		xmlnode4 = xmlnode4->next;
	      }
	    }
	    xmlnode3 = xmlnode3->next;
	  }
	}
	xmlnode2 = xmlnode2->next;
      }
    }
    xmlnode1 = xmlnode1->next;
  }

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  if (cpu.unit != "specint2000seconds")
    cerr << "Reqested unit: " << cpu.unit << " "
	 << "Expected unit: specint2000seconds" << endl;

  int njob = 0;

  stringstream stream, dummy;
  streambuf * sbo = cout.rdbuf(stream.rdbuf());
  streambuf * sbe = cerr.rdbuf(dummy.rdbuf());
  vector <string> e;
  ngstatxx (e, e, e, e, e, false, e, true, true, 20, 0, true);
  cerr.rdbuf(sbe);
  cout.rdbuf(sbo);

  string s = stream.str();

  while (s.size()>0) {
    string line;
    int clusterfreq;
    int clusterncpu;
    int clusterucpu;
    int clusternjob;
    int clusterqjob;
    int clustermem;
    int queuefreq;
    int queuencpu;
    int queueucpu;
    int queuenjob;
    int queueqjob;
    int queuemem;
    int maxtime;
    int mintime;
    int maxuserrun;
    int userfree;
    int userqueue;
    long long userdisk;
    vector <string> rte;

    GetLine (s, line);

    if (line.substr (0, 7) == "Cluster") {
      clusterfreq = 1000;
      clusterncpu = UNDEFINED;
      clusterucpu = UNDEFINED;
      clusternjob = UNDEFINED;
      clusterqjob = UNDEFINED;
      clustermem = UNDEFINED;
      rte.clear();
      while (s.size()>0) {
	GetLine (s, line);
	if (line.substr (0, 2) != "  ") break;
	int cpos = line.find (':', 2);
	if (cpos != string::npos) {
	  string key = line.substr (2, cpos - 2);
	  string val;
	  if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	  if (key == "Installed runtime environments") {
	    while (s.size()>0) {
	      GetLine (s, line);
	      if (line.substr (0, 4) != "    ") break;
	      rte.push_back (line.substr (4));
	    }
	    if (line.substr (0, 2) != "  ") break;
	    cpos = line.find (':', 2);
	    if (cpos != string::npos) {
	      key = line.substr (2, cpos - 2);
	      if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	    }
	  }
	  if (key == "CPU") {
	    int pos = val.find (" MHz");
	    if (pos != string::npos) {
	      int numpos = val.rfind (' ', pos - 1) + 1;
	      clusterfreq = atoi (val.substr (numpos, pos - numpos).c_str());
	    }
	  }
	  if (key == "Number of CPUs")
	    clusterncpu = atoi (val.c_str());
	  if (key == "Number of used CPUs")
	    clusterucpu = atoi (val.c_str());
	  if (key == "Number of jobs")
	    clusternjob = atoi (val.c_str());
	  if (key == "Number of queued jobs")
	    clusterqjob = atoi (val.c_str());
	  if (key == "Memory on each node")
	    clustermem = atoi (val.c_str());
	}
      }
    }

    if (line.substr (0, 5) == "Queue") {
      queuefreq = clusterfreq;
      queuencpu = clusterncpu;
      queueucpu = clusterucpu;
      queuenjob = clusternjob;
      queueqjob = clusterqjob;
      queuemem = clustermem;
      mintime = UNDEFINED;
      maxtime = UNDEFINED;
      maxuserrun = UNDEFINED;
      userfree = UNDEFINED;
      userqueue = UNDEFINED;
      userdisk = UNDEFINED;
      while (s.size()>0) {
	GetLine (s, line);
	if (line.substr (0, 2) != "  ") break;
	int cpos = line.find (':', 2);
	if (cpos != string::npos) {
	  string key = line.substr (2, cpos - 2);
	  string val;
	  if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	  if (key == "User number of free CPUs") {
	    while (s.size()>0) {
	      GetLine (s, line);
	      if (line.substr (0, 4) != "    ") break;
	      if (line.find ("None") != string::npos)
		userfree = 0;
	      else if (line.find ("for infinite time") != string::npos) {
		int ppos = line.find (" processor");
		if (ppos != string::npos) {
		  int free = atoi (line.substr (4, ppos - 4).c_str());
		  if (userfree == UNDEFINED || userfree < free)
		    userfree = free;
		}
	      }
	      else {
		int fpos = line.find ("for ");
		if (fpos != string::npos && Minutes (line.substr (fpos + 4)) >
		    atoi (cpu.count.c_str()) / SI2000PERMHZ / queuefreq) {
		  int ppos = line.find (" processor");
		  if (ppos != string::npos) {
		    int free = atoi (line.substr (4, ppos - 4).c_str());
		    if (userfree == UNDEFINED || userfree < free)
		      userfree = free;
		  }
		}
	      }
	    }
	    if (line.substr (0, 2) != "  ") break;
	    cpos = line.find (':', 2);
	    if (cpos != string::npos) {
	      key = line.substr (2, cpos - 2);
	      if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	    }
	  }
	  if (key == "CPU") {
	    int pos = val.find (" MHz");
	    if (pos != string::npos) {
	      int numpos = val.rfind (' ', pos - 1) + 1;
	      queuefreq = atoi (val.substr (numpos, pos - numpos).c_str());
	    }
	  }
	  if (key == "Number of CPUs")
	    queuencpu = atoi (val.c_str());
	  if (key == "Number of used CPUs")
	    queueucpu = atoi (val.c_str());
	  if (key == "Number of jobs")
	    queuenjob = atoi (val.c_str());
	  if (key == "Number of queued jobs")
	    queueqjob = atoi (val.c_str());
	  if (key == "Memory on each node")
	    queuemem = atoi (val.c_str());
	  if (key == "Min CPU time")
	    mintime = Minutes (val);
	  if (key == "Max CPU time")
	    maxtime = Minutes (val);
	  if (key == "Max number of running jobs per local user")
	    maxuserrun = atoi (val.c_str());
	  if (key == "User queue length")
	    userqueue = atoi (val.c_str());
	  if (key == "User disk space")
	    userdisk = atoll (val.c_str());
	}
      }

      if (!transUses.empty()) {
	bool found = false;
	for (vector <string>::iterator vsi = rte.begin(); !found && vsi != rte.end(); vsi++)
	  if (strcasecmp (vsi->c_str(), transUses.c_str()) == 0) found = true;
	if (!found) continue;
      }

      long long rdisk = atoll (disk.count.c_str());
      if      (disk.unit == "KB") rdisk *= 1024;
      else if (disk.unit == "MB") rdisk *= 1024 * 1024;
      else if (disk.unit == "GB") rdisk *= 1024 * 1024 * 1024;

      long long rmem = atoll (ram.count.c_str());
      if      (ram.unit == "KB") rmem *= 1024;
      else if (ram.unit == "MB") rmem *= 1024 * 1024;
      else if (ram.unit == "GB") rmem *= 1024 * 1024 * 1024;

      if (rmem <= queuemem && userfree != UNDEFINED &&
	  (mintime == UNDEFINED ||
	   atoi (cpu.count.c_str()) / SI2000PERMHZ / queuefreq >= mintime) &&
	  (maxtime == UNDEFINED ||
	   atoi (cpu.count.c_str()) / SI2000PERMHZ / queuefreq <= maxtime) &&
	  queueucpu + queueqjob < 2 * queuencpu)
	njob += std::min (userdisk / rdisk, (long long) 2 * queuencpu - queueucpu - queueqjob);
    }
  }

  stringstream conv;
  conv << njob;
  string njobs;
  njobs = conv.str();

  xmlDocPtr doc = xmlNewDoc (BAD_CAST "1.0");

  xmlCreateIntSubset (doc, BAD_CAST "windmill", BAD_CAST "SYSTEM",
		      BAD_CAST "exe_numJobsWanted.dtd");

  xmlNodePtr root = xmlNewDocNode (doc, NULL, BAD_CAST "windmill", NULL);
  xmlDocSetRootElement (doc, root);
  xmlNewProp (root, BAD_CAST "version", BAD_CAST "0.5");
  xmlNewProp (root, BAD_CAST "user", BAD_CAST "executor");
  xmlNewProp (root, BAD_CAST "type", BAD_CAST "respond");

  xmlNodePtr njw = xmlNewChild (root, NULL, BAD_CAST "numJobsWanted", NULL);
  xmlNodePtr avr = xmlNewChild (njw, NULL, BAD_CAST "availableResources", NULL);
  xmlNewChild (avr, NULL, BAD_CAST "jobCount", BAD_CAST njobs.c_str());
  xmlNodePtr cpun = xmlNewChild (avr, NULL, BAD_CAST "cpuMax", NULL);
  xmlNewChild (cpun, NULL, BAD_CAST "count", BAD_CAST cpu.count.c_str());
  xmlNewChild (cpun, NULL, BAD_CAST "unit", BAD_CAST cpu.unit.c_str());

  xmlChar * x;
  int len;

  xmlDocDumpFormatMemoryEnc (doc, &x, &len, "utf-8", 0);

  xmlFreeDoc (doc);

  cout << x << endl;

  return (char *) x;
}

char * executeJobs (char * xml) {

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);
  xmlNodePtr xmlnode = xmlroot->xmlChildrenNode;
  vector<jobDef> vJobDef;

  while(xmlnode) {
    if(strcmp ((const char*)xmlnode->name, "executeJobs") == 0) {
      xmlNodePtr xmlnode1 = xmlnode->xmlChildrenNode;

      while(xmlnode1) {
	if(strcmp ((const char*) xmlnode1->name, "jobDef") == 0) {
	  xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;

	  jobDef job;

	  while(xmlnode2) {
	    if(strcmp ((const char*)xmlnode2->name, "jobDefID") == 0) {
	      if(xmlnode2->xmlChildrenNode) job.jobDefID = (const char*)xmlnode2->xmlChildrenNode->content;
	    }
	    if(strcmp ((const char*)xmlnode2->name, "jobName") == 0) {
	      if(xmlnode2->xmlChildrenNode) job.jobName = (const char*)xmlnode2->xmlChildrenNode->content;
	    }
	    if(strcmp ((const char*)xmlnode2->name, "transImplementation") == 0) {
	      if(xmlnode2->xmlChildrenNode) job.transImpl = (char*)xmlnode2->xmlChildrenNode->content;
	    }
	    if(strcmp ((const char*)xmlnode2->name, "transUses") == 0) {
	      if(xmlnode2->xmlChildrenNode) job.transUses = (char*)xmlnode2->xmlChildrenNode->content;
	    }

	    if(strcmp ((const char*)xmlnode2->name, "expectedResources") == 0) {
	      xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	      while(xmlnode3) {
		if(strcmp((const char*)xmlnode3->name, "cpuConsumption") == 0) {
		  xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
		  while(xmlnode4) {
		    if (strcmp ((const char*)xmlnode4->name, "count") == 0)
		      if(xmlnode4->xmlChildrenNode) job.CPU.count = (const char*)xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*)xmlnode4->name, "unit") == 0)
		      if(xmlnode4->xmlChildrenNode) job.CPU.unit = (const char*)xmlnode4->xmlChildrenNode->content;
		    xmlnode4 = xmlnode4->next;
		  }
		}
		if(strcmp ((const char*)xmlnode3->name, "diskConsumption") == 0) {
		  xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
		  while (xmlnode4) {
		    if (strcmp ((const char*)xmlnode4->name, "count") == 0)
		      if(xmlnode4->xmlChildrenNode) job.disk.count = (const char*)xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*)xmlnode4->name, "unit") == 0)
		      if(xmlnode4->xmlChildrenNode) job.disk.unit = (const char*)xmlnode4->xmlChildrenNode->content;
		    xmlnode4 = xmlnode4->next;
		  }
		}
		if(strcmp ((const char*)xmlnode3->name, "minimumRAM") == 0) {
		  xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
		  while (xmlnode4) {
		    if (strcmp ((const char*)xmlnode4->name, "count") == 0)
		      if(xmlnode4->xmlChildrenNode) job.ram.count = (const char*)xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*)xmlnode4->name, "unit") == 0)
		      if(xmlnode4->xmlChildrenNode) job.ram.unit = (const char*)xmlnode4->xmlChildrenNode->content;
		    xmlnode4 = xmlnode4->next;
		  }
		}
		if(strcmp ((const char*)xmlnode3->name, "ipConnectivity") == 0) {
		  if(xmlnode3->xmlChildrenNode) job.ipconnectivity = (char*)xmlnode3->xmlChildrenNode->content;
		}
		xmlnode3 = xmlnode3->next;
	      }
	    }

	    if(strcmp((const char*)xmlnode2->name, "jobPars") == 0) {
	      xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	      while(xmlnode3) {
		if(strcmp((const char*)xmlnode3->name, "actualPar") == 0) {
		  xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;

		  actualPar pars;
		  while(xmlnode4) {
		    if (strcmp ((const char*)xmlnode4->name, "name") == 0)
		      if(xmlnode4->xmlChildrenNode) pars.name = (const char*)xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*)xmlnode4->name, "value") == 0)
		      if(xmlnode4->xmlChildrenNode) pars.value = (const char*)xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*)xmlnode4->name, "position") == 0)
		      if(xmlnode4->xmlChildrenNode) pars.position = atoi( (const char*)xmlnode4->xmlChildrenNode->content );
		    if (strcmp ((const char*)xmlnode4->name, "type") == 0)
		      if(xmlnode4->xmlChildrenNode) pars.type = (const char*)xmlnode4->xmlChildrenNode->content;
		    xmlnode4 = xmlnode4->next;
		  }
		  job.jobPars.push_back( pars );
		}
		xmlnode3 = xmlnode3->next;
	      }
	    }

	    if( (strcmp ((const char*)xmlnode2->name, "jobInputs") == 0 )||(strcmp( (char*)xmlnode2->name, "jobOutputs" ) == 0)||(strcmp( (char*)xmlnode2->name, "jobLogs" ) == 0) ) {
	      xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	      while(xmlnode3) {
		if(strcmp ((const char*)xmlnode3->name, "fileInfo") == 0) {

		  fileInfo info;
		  xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
		  while(xmlnode4) {
		    if (strcmp ((const char*)xmlnode4->name, "stream") == 0)
		      if(xmlnode4->xmlChildrenNode) info.stream = (const char*)xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*)xmlnode4->name, "LFN") == 0)
		      if(xmlnode4->xmlChildrenNode) info.LFN = (const char*)xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*)xmlnode4->name, "type") == 0)
		      if(xmlnode4->xmlChildrenNode) info.type = (const char*)xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*)xmlnode4->name, "logCol") == 0)
		      if(xmlnode4->xmlChildrenNode) info.logCol = (const char*)xmlnode4->xmlChildrenNode->content;

		    if (strcmp ((const char*)xmlnode4->name, "SEList") == 0) {
		      xmlNodePtr xmlnode5 = xmlnode4->xmlChildrenNode;
		      while(xmlnode5) {
			if(strcmp ((const char*)xmlnode5->name, "SE") == 0)
			  info.SEList.push_back( (const char*)xmlnode5->xmlChildrenNode->content );
			xmlnode5 = xmlnode5->next;
		      }
		    }

		    if (strcmp ((const char*)xmlnode4->name, "turlList") == 0) {
		      xmlNodePtr xmlnode5 = xmlnode4->xmlChildrenNode;
		      while(xmlnode5) {
			if( strcmp ((const char*)xmlnode5->name, "turl") == 0)
			  info.turlList.push_back((const char*)xmlnode5->xmlChildrenNode->content );
			xmlnode5 = xmlnode5->next;
		      }
		    }

		    xmlnode4 = xmlnode4->next;
		  }

		  if(strcmp ((const char*)xmlnode2->name,"jobInputs") == 0) job.jobInputs.push_back( info );
		  if(strcmp ((const char*)xmlnode2->name,"jobOutputs") == 0) job.jobOutputs.push_back( info );
		  if(strcmp ((const char*)xmlnode2->name,"jobLogs") == 0) job.jobLogs.push_back( info );
		}
		xmlnode3 = xmlnode3->next;
	      }
	    }
	    xmlnode2 = xmlnode2->next;
	  }

	  vJobDef.push_back( job );
	}

	xmlnode1 = xmlnode1->next;
      }
    }

    xmlnode = xmlnode->next;
  }

  // constructing xrslstring
  vector<string> xrslstrings;

  for( unsigned int i=0; i<vJobDef.size(); i++) {

    string xrslstring = "&";
    jobDef job = vJobDef[i];

    if( job.transImpl.size()>0 )
      xrslstring = xrslstring + "(executable=\"" + job.transImpl + "\")";

    int maxarg = 0;
    for( unsigned int j=0; j<job.jobPars.size(); j++ ) {
      if( maxarg<job.jobPars[j].position ) maxarg=job.jobPars[j].position;
    }
    vector<string> arguments( maxarg );
    for( unsigned int j=0; j<job.jobPars.size(); j++ ) {
      arguments[job.jobPars[j].position - 1] = job.jobPars[j].value;
    }
    if( arguments.size()>0 ) {
      xrslstring += "(arguments=";
      for( unsigned int j=0; j<maxarg; j++ ) {
	xrslstring = xrslstring + arguments[j] + " ";
      }
      xrslstring.erase( xrslstring.size()-1 );
      xrslstring += ")";
    }

    string stdout;
    string stderr;
    string join;
    if( job.jobLogs.size()==1 )
      join = "yes";
    for( int i=0; i<job.jobLogs.size(); i++ ) {
      if( strcmp( job.jobLogs[i].stream.c_str(), "stdout" ) == 0 )
        stdout= job.jobLogs[i].LFN;
      if( strcmp( job.jobLogs[i].stream.c_str(), "stderr" ) == 0 )
        stderr = job.jobLogs[i].LFN;
      if( strcmp( job.jobLogs[i].stream.c_str(), "stdboth") == 0 ) {
        stdout = job.jobLogs[i].LFN;
        join = "yes";
      }
    }

    if( stdout.size()>0 )
      xrslstring = xrslstring + "(stdout=\"" + stdout + "\")";
    if( stderr.size()>0 )
      xrslstring = xrslstring + "(stderr=\"" + stderr + "\")";
    if( join.size()>0 )
      xrslstring = xrslstring + "(join=yes)";

    if( job.transUses.size()>0 )
      xrslstring = xrslstring + "(runTimeEnvironment=\"" + job.transUses + "\")";
    if( (job.disk.count.size()>0) && (job.disk.unit.size()>0) ) {
      string size;
      stringstream stream;
      if( strncmp( job.disk.unit.c_str(),"KB",2 ) == 0 ) {
	int newsize = 1 + atoll( job.disk.count.c_str() )/1000;
	stream << newsize;
	size = stream.str();
      } else if( strncmp( job.disk.unit.c_str(),"GB",2 ) == 0 ) {
	long newsize = atoll( job.disk.count.c_str() )*1000;
	stream << newsize;
	size = stream.str();
      } else if( strncmp( job.disk.unit.c_str(),"MB",2 ) == 0 ) {
	size = job.disk.count;
      } else {
	cerr << "Unsupported disk-unit! Assuming MB." << endl;
	size = job.disk.count;
      }
      xrslstring = xrslstring + "(disk=" + size + ")";
    }

    if( (job.ram.count.size()>0) && (job.ram.unit.size()>0) ) {
      string size;
      stringstream stream;
      if( strncmp( job.ram.unit.c_str(),"KB",2 ) == 0 ) {
	int newsize = 1 + atoll( job.ram.count.c_str() )/1000;
	stream << newsize;
	size = stream.str();
      } else if( strncmp( job.ram.unit.c_str(),"GB",2 ) == 0 ) {
	long newsize = atoll( job.ram.count.c_str() )*1000;
	stream << newsize;
	size = stream.str();
      } else if( strncmp( job.ram.unit.c_str(),"MB",2 ) == 0 ) {
	size = job.ram.count;
      } else {
	cerr << "Unsupported ram-unit! Assuming MB." << endl;
	size = job.ram.count;
      }
      xrslstring = xrslstring + "(memory=" + size + ")";
    }

    if( strncmp( job.ipconnectivity.c_str(), "yes", 3 ) == 0 ) {
      xrslstring = xrslstring + "(nodeAccess=outbound)";
    }

    if( (job.CPU.count.size()>0) && (job.CPU.unit.size()>0) ) {
      xrslstring = xrslstring + "(GridTime=\"";
      stringstream timestring;
      timestring << (int) (atoll(job.CPU.count.c_str()) / (SI2000PERMHZ*2800));
      xrslstring = xrslstring + timestring.str() + "\")";
    }
    if( (job.jobName.size()>0)||(job.jobDefID.size()>0) )
      xrslstring = xrslstring + "(jobName=\"" + job.jobName + "\")";
    if( job.jobInputs.size()>0 ) {
      xrslstring = xrslstring + "(inputFiles=";
      for( unsigned int j=0; j<job.jobInputs.size(); j++ ) {
	if( job.jobInputs[j].turlList.size()>0 ) {
	  xrslstring = xrslstring + "(\"" + job.jobInputs[j].LFN + "\" \"" + job.jobInputs[j].turlList[0]+ "\")";
	}
      }
      xrslstring = xrslstring + ")";
    }

    // modify this to use the RLS server
    if( job.jobOutputs.size()>0 ) {
      xrslstring = xrslstring + "(outputFiles=";
      for( unsigned int j=0; j<job.jobOutputs.size(); j++ ) {
	if( job.jobOutputs[j].SEList.size()>0 ) {
	  xrslstring = xrslstring + "(\"" + job.jobOutputs[j].LFN + "\" \"rc://";
	  for( unsigned int k=0; k<job.jobOutputs[j].SEList.size(); k++ ) {
	    xrslstring = xrslstring + job.jobOutputs[j].SEList[k];
	    if( k != job.jobOutputs[j].SEList.size()-1 ) xrslstring = xrslstring + "|";
	  }
	  xrslstring = xrslstring + "@grid.uio.no/lc=TestCollection,rc=NorduGrid,dc=nordugrid,dc=org/" + job.jobOutputs[j].logCol + "/" + job.jobOutputs[j].LFN + "\")";
	}
      }
      if( job.jobLogs.size()==0 ) xrslstring = xrslstring + ")";
    }

    if( job.jobLogs.size()>0 ) {
      if( job.jobOutputs.size()==0 ) xrslstring = xrslstring + "(outputFiles=";
      for( unsigned int j=0; j<job.jobLogs.size(); j++ ) {
        if( job.jobLogs[j].SEList.size()>0 ) {
          xrslstring = xrslstring + "(\"" + job.jobLogs[j].LFN + "\" \"rc://";
          for( unsigned int k=0; k<job.jobLogs[j].SEList.size(); k++ ) {
            xrslstring = xrslstring + job.jobLogs[j].SEList[k];
            if( k != job.jobLogs[j].SEList.size()-1 ) xrslstring = xrslstring +
"|";
          }
          xrslstring = xrslstring + "@grid.uio.no/lc=TestCollection,rc=NorduGrid,dc=nordugrid,dc=org/" + job.jobLogs[j].logCol + "/" + job.jobLogs[j].LFN + "\")";
        }
      }
      xrslstring = xrslstring + ")";
    }
    
    xrslstrings.push_back( xrslstring );
  }

  stringstream s, dummy;
  streambuf * sbo = cout.rdbuf(s.rdbuf());
  streambuf * sbe = cerr.rdbuf(dummy.rdbuf());
  vector <string> e;
  string f;
  ngsubxx (e, xrslstrings, e, e, e, f, false, false, 20, 2, true);
  cerr.rdbuf(sbe);
  cout.rdbuf(sbo);

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  return NULL;
}

char * getExecutorData (char * xml) {

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

  xmlNodePtr xmlnode1 = xmlroot->xmlChildrenNode;

  vector <executorData> exdata;

  while (xmlnode1) {
    if (strcmp ((const char *) xmlnode1->name, "getExecutorData") == 0) {
      xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;
      while (xmlnode2) {
	if (strcmp ((const char *) xmlnode2->name, "jobDef") == 0) {
	  xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	  executorData jobDef;
	  while (xmlnode3) {
	    if (strcmp ((const char *) xmlnode3->name, "jobDefID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobDefID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobName") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobName =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    xmlnode3 = xmlnode3->next;
	  }
	  exdata.push_back (jobDef);
	}
	xmlnode2 = xmlnode2->next;
      }
    }
    xmlnode1 = xmlnode1->next;
  }

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  string filename = GetEnv ("HOME");
  filename.append ("/.ngjobs");
  ifstream ngjobs (filename.c_str());

  for (vector <executorData>::iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++) {
    ngjobs.clear();
    ngjobs.seekg (ios::beg);
    string jobidname;
    while (getline (ngjobs, jobidname)) {
      int pos = jobidname.find('#');
      string jobid = jobidname.substr (0, pos);
      string jobname = jobidname.substr (pos + 1);
      if (jobname == vedi->jobName)
	vedi->jobExeID = jobid;
    }
  }

  ngjobs.close();

  xmlDocPtr doc = xmlNewDoc (BAD_CAST "1.0");

  xmlCreateIntSubset (doc, BAD_CAST "windmill", BAD_CAST "SYSTEM",
		      BAD_CAST "exe_getExecutorData.dtd");

  xmlNodePtr root = xmlNewDocNode (doc, NULL, BAD_CAST "windmill", NULL);
  xmlDocSetRootElement (doc, root);
  xmlNewProp (root, BAD_CAST "version", BAD_CAST "0.5");
  xmlNewProp (root, BAD_CAST "user", BAD_CAST "executor");
  xmlNewProp (root, BAD_CAST "type", BAD_CAST "respond");
  xmlNodePtr ged = xmlNewChild (root, NULL, BAD_CAST "getExecutorData", NULL);

  for (vector <executorData>::const_iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++) {
    xmlNodePtr jed = xmlNewChild (ged, NULL, BAD_CAST "jobExeData", NULL);
    xmlNewChild (jed, NULL, BAD_CAST "jobDefID",
		 BAD_CAST vedi->jobDefID.c_str());
    xmlNewChild (jed, NULL, BAD_CAST "jobName",
		 BAD_CAST vedi->jobName.c_str());
    if (!vedi->jobExeID.empty())
      xmlNewChild (jed, NULL, BAD_CAST "jobExeID",
		   BAD_CAST vedi->jobExeID.c_str());
    xmlNewChild (jed, NULL, BAD_CAST "submitStatus",
		 BAD_CAST (vedi->jobExeID.empty() ? "unknown" : "submitted"));
  }

  xmlChar * x;
  int len;

  xmlDocDumpFormatMemoryEnc (doc, &x, &len, "utf-8", 0);

  xmlFreeDoc (doc);

  cout << x << endl;

  return (char *) x;
}

char * getStatus (char * xml) {

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

  xmlNodePtr xmlnode1 = xmlroot->xmlChildrenNode;

  vector <jobExeData> exdata;

  while (xmlnode1) {
    if (strcmp ((const char *) xmlnode1->name, "getStatus") == 0) {
      xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;
      while (xmlnode2) {
	if (strcmp ((const char *) xmlnode2->name, "jobExeData") == 0) {
	  xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	  jobExeData jobDef;
	  while (xmlnode3) {
	    if (strcmp ((const char *) xmlnode3->name, "jobDefID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobDefID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobName") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobName =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobExeID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobExeID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "infoExecutor") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.infoExecutor =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    xmlnode3 = xmlnode3->next;
	  }
	  exdata.push_back (jobDef);
	}
	xmlnode2 = xmlnode2->next;
      }
    }
    xmlnode1 = xmlnode1->next;
  }

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  vector <string> jobs;

  for (vector <jobExeData>::iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++)
    jobs.push_back (vedi->jobExeID);

  stringstream stream, dummy;
  streambuf * sbo = cout.rdbuf(stream.rdbuf());
  streambuf * sbe = cerr.rdbuf(dummy.rdbuf());
  vector <string> e;
  ngstatxx (jobs, e, e, e, e, false, e, false, true, 20, 0, true);
  cerr.rdbuf(sbe);
  cout.rdbuf(sbo);

  map <string, string> status;
  map <string, string> error;
  map <string, string> submitted;
  map <string, string> gmlog;
  map <string, string> host;

  string s = stream.str();

  while (s.size()>0) {
    string line;

    GetLine (s, line);

    if (line.substr (0, 3) == "Job") {
      string jobid = line.substr (4);
      while (s.size()>0) {
	GetLine (s, line);
	if (line.substr (0, 2) != "  ") break;
	int cpos = line.find (':', 2);
	if (cpos != string::npos) {
	  string key = line.substr (2, cpos - 2);
	  string val;
	  if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	  if (key == "Execution nodes") {
	    while (s.size()>0) {
	      GetLine (s, line);
	      if (line.substr (0, 4) != "    ") break;
	      host [jobid] = line.substr (4);
	      // Need to be improved for parallel jobs, but for Atlas do we need it?
	    }
	    if (line.substr (0, 2) != "  ") break;
	    cpos = line.find (':', 2);
	    if (cpos != string::npos) {
	      key = line.substr (2, cpos - 2);
	      if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	    }
	  }
	  if (key == "Status")
	    status [jobid] = val;
	  if (key == "Error")
	    error [jobid] = val;
	  if (key == "Submitted")
	    submitted [jobid] = val;
	  if (key == "Grid manager log directory")
	    gmlog [jobid] = val;
	}
      }
    }
  }

  map <string, string> xrsl;

  for (map <string, string>::iterator mssi = gmlog.begin();
       mssi != gmlog.end(); mssi++) {
    string xrslurl = mssi->second + "/description";
    char * tempname = "/tmp/xrsl.XXXXXX";
    mktemp (tempname);
    char ** args;
    args [0] = "ngcopy";
    args [1] = (char *) xrslurl.c_str();
    args [2] = (char *) ((string) "file://" + tempname).c_str();
    ngcopy (2, args);
    ifstream xrslfile (tempname);
    xrslfile.seekg (0, ios::end);
    int length = xrslfile.tellg();
    xrslfile.seekg (0, ios::beg);
    char * buffer = new char [length + 1];
    xrslfile.read (buffer, length);
    buffer [length] = 0;
    xrslfile.close();
    xrsl [mssi->first] = buffer;
    delete buffer;
    remove (tempname);
  }

  xmlDocPtr doc = xmlNewDoc (BAD_CAST "1.0");

  xmlCreateIntSubset (doc, BAD_CAST "windmill", BAD_CAST "SYSTEM",
		      BAD_CAST "exe_getStatus.dtd");

  xmlNodePtr root = xmlNewDocNode (doc, NULL, BAD_CAST "windmill", NULL);
  xmlDocSetRootElement (doc, root);
  xmlNewProp (root, BAD_CAST "version", BAD_CAST "0.5");
  xmlNewProp (root, BAD_CAST "user", BAD_CAST "executor");
  xmlNewProp (root, BAD_CAST "type", BAD_CAST "respond");
  xmlNodePtr ged = xmlNewChild (root, NULL, BAD_CAST "getStatus", NULL);

  globus_module_activate (GLOBUS_RSL_MODULE);

  for (vector <jobExeData>::const_iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++) {
    xmlNodePtr jed = xmlNewChild (ged, NULL, BAD_CAST "jobExeData", NULL);
    xmlNewChild (jed, NULL, BAD_CAST "jobDefID",
		 BAD_CAST vedi->jobDefID.c_str());
    xmlNewChild (jed, NULL, BAD_CAST "jobName",
		 BAD_CAST vedi->jobName.c_str());
    xmlNewChild (jed, NULL, BAD_CAST "jobExeID",
		 BAD_CAST vedi->jobExeID.c_str());
    if (!vedi->infoExecutor.empty())
      xmlNewChild (jed, NULL, BAD_CAST "infoExecutor",
		   BAD_CAST vedi->infoExecutor.c_str());

    string jobstatus = "unknown";
    if (status [vedi->jobExeID] == "ACCEPTED") jobstatus = "pending";
    if (status [vedi->jobExeID] == "PREPARING" ||
	status [vedi->jobExeID].substr (0, 6) == "INLRMS" ||
	status [vedi->jobExeID] == "FINISHING") jobstatus = "running";
    if (status [vedi->jobExeID].substr (0, 8) == "FINISHED") jobstatus = "finished";
    if (!error [vedi->jobExeID].empty()) jobstatus = "failed";
    xmlNewChild (jed, NULL, BAD_CAST "jobStatus", BAD_CAST jobstatus.c_str());

    xmlNewChild (jed, NULL, BAD_CAST "jobNativeStatus",
		 BAD_CAST status [vedi->jobExeID].c_str());

    if (!error [vedi->jobExeID].empty())
      xmlNewChild (jed, NULL, BAD_CAST "valdationDiagnostic",
		   BAD_CAST error [vedi->jobExeID].c_str());

    if (!host [vedi->jobExeID].empty())
      xmlNewChild (jed, NULL, BAD_CAST "host",
		   BAD_CAST host [vedi->jobExeID].c_str());

    int pos1 = error [vedi->jobExeID].find ("JOB FAILURE: Job exit code is ");
    if (pos1 != string::npos) {
      pos1 += 30;
      int pos2 = error [vedi->jobExeID].find (" != 0", pos1);
      xmlNewChild (jed, NULL, BAD_CAST "jobExitCode",
		   BAD_CAST error [vedi->jobExeID].substr (pos1, pos2 - pos1).c_str());
    }
    else
      xmlNewChild (jed, NULL, BAD_CAST "jobExitCode", BAD_CAST "0");

    if (status [vedi->jobExeID].substr (0, 8) == "FINISHED")
      xmlNewChild (jed, NULL, BAD_CAST "endTime",
		   BAD_CAST (status [vedi->jobExeID].substr (9).c_str()));

    if (xrsl.find (vedi->jobExeID) != xrsl.end()) {

      xmlNodePtr opf = xmlNewChild (jed, NULL, BAD_CAST "jobOutputs", NULL);

      globus_rsl_t * axrsl =
	globus_rsl_parse ((char *) xrsl [vedi->jobExeID].c_str());

      globus_rsl_t * outputfilesrelation = NULL;
      FindRelation ("outputfiles", &outputfilesrelation, axrsl);

      if (outputfilesrelation) {

	globus_rsl_value_t * sequence =
	  globus_rsl_relation_get_value_sequence (outputfilesrelation);
	if (!globus_rsl_value_is_sequence (sequence))
	  cerr << "Error: XRSL syntax error in attribute \"outputfiles\"" << endl;

	globus_list_t * list1 =
	  globus_rsl_value_sequence_get_value_list (sequence);

	while (!globus_list_empty (list1)) {

	  globus_rsl_value_t * value1 =
	    (globus_rsl_value_t *) globus_list_first (list1);
	  if (!globus_rsl_value_is_sequence (value1))
	    cerr << "Error: XRSL syntax error in attribute \"outputfiles\"" << endl;

	  globus_list_t * list2 =
	    globus_rsl_value_sequence_get_value_list (value1);

	  int paircnt = 0;
	  string lfn;
	  string pfn;

	  while (!globus_list_empty (list2)) {

	    globus_rsl_value_t * value2 =
	      (globus_rsl_value_t *) globus_list_first (list2);

	    switch (paircnt) {

	    case 0:
	      if (globus_rsl_value_is_literal (value2))
		lfn = globus_rsl_value_literal_get_string (value2);
	      break;

	    case 1:
	      if (globus_rsl_value_is_literal (value2)) {
		pfn = globus_rsl_value_literal_get_string (value2);
		if (pfn.empty()) pfn = lfn;
	      }
	      break;
	    }

	    paircnt++;
	    list2 = globus_list_rest (list2);
	  }
	  if (paircnt != 2)
	    cerr << "Error: XRSL syntax error in attribute \"outputfiles\"" << endl;

	  xmlNodePtr fin = xmlNewChild (opf, NULL, BAD_CAST "fileInfo", NULL);
	  xmlNewChild (fin, NULL, BAD_CAST "LFN", BAD_CAST lfn.c_str());
	  xmlNodePtr tul = xmlNewChild (fin, NULL, BAD_CAST "turlList", NULL);
	  xmlNewChild (tul, NULL, BAD_CAST "turl", BAD_CAST pfn.c_str());

	  list1 = globus_list_rest (list1);
	}
      }
    }
  }

  globus_module_deactivate (GLOBUS_RSL_MODULE);

  xmlChar * x;
  int len;

  xmlDocDumpFormatMemoryEnc (doc, &x, &len, "utf-8", 0);

  xmlFreeDoc (doc);

  cout << x << endl;

  return (char *) x;
}

char * fixJobs (char * xml) {

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

  xmlNodePtr xmlnode1 = xmlroot->xmlChildrenNode;

  vector <jobExeData> exdata;

  while (xmlnode1) {
    if (strcmp ((const char *) xmlnode1->name, "killJob") == 0) {
      xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;
      while (xmlnode2) {
	if (strcmp ((const char *) xmlnode2->name, "jobExeData") == 0) {
	  xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	  jobExeData jobDef;
	  while (xmlnode3) {
	    if (strcmp ((const char *) xmlnode3->name, "jobDefID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobDefID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobName") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobName =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobExeID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobExeID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "infoExecutor") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.infoExecutor =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "action") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.action =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    xmlnode3 = xmlnode3->next;
	  }
	  exdata.push_back (jobDef);
	}
	xmlnode2 = xmlnode2->next;
      }
    }
    xmlnode1 = xmlnode1->next;
  }

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  vector <string> cleanids;
  vector <string> restageids;
  vector <string> alertids;
  for (vector <jobExeData>::const_iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++) {
    if (vedi->action == "clean")
      cleanids.push_back (vedi->jobExeID);
    if (vedi->action == "restage")
      restageids.push_back (vedi->jobExeID);
    if (vedi->action == "alert")
      alertids.push_back (vedi->jobExeID);
  }

  stringstream s, dummy;
  streambuf * sbo = cout.rdbuf(s.rdbuf());
  streambuf * sbe = cerr.rdbuf(dummy.rdbuf());
  vector <string> e;
  string f;
  if (!cleanids.empty())
    ngcleanxx (cleanids, e, e, e, e, false, false, 20, 0, true);
  if (!restageids.empty())
    ngresubxx (restageids, e, e, e, e, false, e, e, e, f, false, false, false,
	       20, 0, true);
  if (!alertids.empty())
    ngrenewxx (alertids, e, e, e, e, false, 20, 0, true);
  cerr.rdbuf(sbe);
  cout.rdbuf(sbo);

  return NULL;
}

char * killJobs (char * xml) {

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

  xmlNodePtr xmlnode1 = xmlroot->xmlChildrenNode;

  vector <jobExeData> exdata;

  while (xmlnode1) {
    if (strcmp ((const char *) xmlnode1->name, "killJob") == 0) {
      xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;
      while (xmlnode2) {
	if (strcmp ((const char *) xmlnode2->name, "jobExeData") == 0) {
	  xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	  jobExeData jobDef;
	  while (xmlnode3) {
	    if (strcmp ((const char *) xmlnode3->name, "jobDefID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobDefID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobName") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobName =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobExeID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobExeID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "infoExecutor") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.infoExecutor =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    xmlnode3 = xmlnode3->next;
	  }
	  exdata.push_back (jobDef);
	}
	xmlnode2 = xmlnode2->next;
      }
    }
    xmlnode1 = xmlnode1->next;
  }

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  vector <string> jobids;
  for (vector <jobExeData>::const_iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++)
    jobids.push_back (vedi->jobExeID);

  stringstream s, dummy;
  streambuf * sbo = cout.rdbuf(s.rdbuf());
  streambuf * sbe = cerr.rdbuf(dummy.rdbuf());
  vector <string> e;
  ngkillxx (jobids, e, e, e, e, false, false, 20, 0, true);
  cerr.rdbuf(sbe);
  cout.rdbuf(sbo);

  return NULL;
}
